/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <iostream>
#include <string>
#include <cstdlib>
#include <algorithm>
#include <boost/regex.hpp>
#include <boost/timer.hpp>
#include <pthread.h>
#include "DBTable.h"
#include "SAColorSetSize.h"
#include "NumericDataValue.h"
#include "TaggedNumericDataValue.h"
#include "Utils.h"
#include "Scores.h"

using namespace std;

//
// output most significant substring pattern pairs.
//

static void print_usage(const char * progname){
  cout << "Usage: " << progname << " [options]" << endl
       << "  -f  db_file : tab separated data file" << endl
       << "  -n  namecol : column # of name of data" << endl
       << "  -v  valcol  : column # of numeric value of data" << endl
       << "  -s  seqcol  : column # of sequence value of data" << endl
       << "  -t  seqcol2 : column # of second sequence data" << endl
       << "  -k  outcnt  : number of results to output" << endl
       << "  -r  regex   : a regex to take substring of sequence data." << endl
       << "  -q  regex2  : second regex to take substring of sequence data." << endl
       << "  -c  combo   : combination to consider (1 to 14 - pattern pair, 0 - single pattern)" << endl
       << "  -l  maxpatl : maximum length for pattern (default 0 = unlimited)" << endl
       << "  -a  n       : maximum # of cpu's to use (default 1)" << endl
       << "  -N          : use naive algorithm O(mN^2)" << endl
       << "  -h  help    : output usage (this message)" << endl;
  return;
}

struct mystrtod : public std::unary_function<double,string> {
  double operator()(const string & s){
    return strtod(s.c_str(), 0);
  }
};

class mysubstr : public std::unary_function<string,string> {
  std::size_t s, l;
  const boost::regex r;
public:
  mysubstr(const string & origin_str) : r(origin_str.c_str()){};
  string operator()(string & s) const {
    boost::smatch what;
    if(boost::regex_search(s, what, r, boost::match_extra)){
      return(what[1]);
    } else {
      return("");
    }
  }
};

template <class DataValue>
struct SAColorSetSizeResultPair {
  double score;
  SASubstring p1;
  SASubstring p2;
  DataValue dv;
  DataValue total;
};

template <class DataValue>
struct CmpPair 
  : public std::binary_function<SAColorSetSizeResultPair<DataValue>,
				SAColorSetSizeResultPair<DataValue>, bool> {
  bool operator()(const SAColorSetSizeResultPair<DataValue> & x,
		  const SAColorSetSizeResultPair<DataValue> & y){
    return ((x.score > y.score) ? true : false);
  }
};

static CmpPair<TaggedNumericDataValue> cmppair = CmpPair<TaggedNumericDataValue>();

struct thread_args {
  size_t id;
  size_t job_begin;
  size_t job_end;
  size_t result_num;
  size_t combo;
  const SuffixArray * sa;
  const SuffixArray * sa2;
  const vector<SASubstring> * p1vec;
  const vector<string> * seqvec;
  const vector<double> * rankvec;
  double vc;
};

void * calc_thread(void * p){
  const thread_args * args = static_cast<thread_args *>(p);
  /* easy access to constant arguments */
  // const SuffixArray & sa = *(args->sa);
  const SuffixArray & sa2 = *(args->sa2);
  const vector<string> & seqvec  = *(args->seqvec);
  const vector<double> & rankvec = *(args->rankvec);
  const char * origtext = sa2.get_string().c_str();
  const size_t result_num = args->result_num;

  /* variables to be used */
  vector<bool> tagvec(seqvec.size());
  vector<TaggedNumericDataValue> correction_factors;
  vector<SAColorSetSizeResultPair<TaggedNumericDataValue> > * resvec
    = new vector<SAColorSetSizeResultPair<TaggedNumericDataValue> >();
  vector<SAColorSetSizeResult<TaggedNumericDataValue> > results;
  const TNDVRankSumTestScore ranksumtest(args->combo, args->vc);
  SAColorSetSizeResultPair<TaggedNumericDataValue> resval;

  results.reserve(result_num);
  resvec->reserve(result_num);

  /* misc variables */
  size_t i, pat1_matches, jobid;
  cout << "Thread: " << args->id << flush << endl;
  for(jobid = args->job_begin; jobid < args->job_end; jobid++){
    if(args->id == 0)
      cout << "Job: " << jobid << " [" << args->job_begin << ", " << args->job_end << ")" << endl;
    SASubstring pat1 = (*args->p1vec)[jobid];
    pat1_matches = 0;
    // set tagvec /////////////////////
    for(i = 0; i < tagvec.size(); i++){
      // http://www.sgi.com/tech/stl/basic_string.html
      // size_type find(const charT* s, size_type pos, size_type n) const
      // Searches for the first n characters of s as a substring of
      // *this, beginning at character pos of *this.
      tagvec[i] = (seqvec[i].find(origtext + sa2.get_suffix(pat1.l),
				  0, pat1.h) != std::string::npos);
      if(tagvec[i]) pat1_matches++;
    }
    ///////////////////////////////////

    // execute CSS algorithm //////////
    const TaggedNumericDataValueFactory fctry(rankvec, tagvec);

    results = SAColorSetSize<TaggedNumericDataValue,
                             TaggedNumericDataValueFactory,
                             TNDVRankSumTestScore>::solve(sa2,
							  fctry,
							  ranksumtest,
							  correction_factors,
							  result_num,
							  false);
    for(i = 0; i < results.size(); i++){
      if(resvec->size() < result_num){
	resval.score = results[i].score;
	resval.p1    = pat1;
	resval.p2    = results[i].node;
	resval.dv    = results[i].dv;
	resval.total = fctry.get_total();
	resvec->push_back(resval);
	push_heap(resvec->begin(), resvec->end(), cmppair);
      } else if(resvec->front().score < results[i].score){
	pop_heap(resvec->begin(), resvec->end(), cmppair);
	resvec->back().score = results[i].score;
	resvec->back().p1    = pat1;
	resvec->back().p2    = results[i].node;
	resvec->back().dv    = results[i].dv;
	resvec->back().total = fctry.get_total();
	push_heap(resvec->begin(), resvec->end(), cmppair);
      }
    }
    ///////////////////////////////////
  }
  cout << "Thread: " << args->id << " finished." << endl;
  return resvec;
}


int main(int argc, char * argv[]){
  int optchar;
  size_t namecol, valcol, seqcol, combo, maxpatl;
  size_t seqcol2;
  size_t i, j, result_num, concurrency;
  string dbfile;
  string regex, regex2;
  bool use_naive = false;

  namecol = 0;
  valcol  = 1;
  seqcol  = seqcol2 = 2;
  combo   = 0; /* default is single pattern */
  maxpatl = 0; /* unlimited */
  result_num = 10;
  concurrency = 1;
  cout << fixed;
  while((optchar = getopt(argc, argv, "hf:n:v:s:c:k:r:l:a:t:q:N")) != -1){
    switch(optchar){
    case 'h': print_usage(argv[0]); exit(0);
    case 'f': dbfile = string(optarg); break;
    case 'n': namecol = atoi(optarg); break;
    case 'v': valcol = atoi(optarg); break;
    case 's': seqcol = atoi(optarg); break;
    case 't': seqcol2 = atoi(optarg); break;
    case 'c': combo = atoi(optarg); break;
    case 'k': result_num = atoi(optarg); break;
    case 'r': regex = string(optarg); break;
    case 'q': regex2 = string(optarg); break;
    case 'l': maxpatl = atoi(optarg); break;
    case 'a': concurrency = atoi(optarg); break;
    case 'N': use_naive = true; break;
    case '?': fprintf(stderr, "Aborting.\n"); exit (1);
    default:
      cerr << "INTERNAL ERROR: getopt returned character code "
	   << (size_t) optchar << ". Aborting." << endl;
      exit (1);
    }
  }

  /********************* BEGIN: Read and process data file *********************/
  /* values to set */
  vector<string> namevec, seqvec, seqvec2;
  vector<double> weightvec, rankvec;

  /* temporary vectors */
  vector<size_t> required_columns, tmpvec;
  vector<string> weightvec_str;

  cout << "Reading: " << dbfile << endl;
  LineProcessor lp("\t");
  DBTable tbl(dbfile, lp);
  
  cout << "Checking columns." << endl;
  required_columns.push_back(namecol); required_columns.push_back(valcol); required_columns.push_back(seqcol); required_columns.push_back(seqcol2);

  // exclude rows with invalid entries in required_columns
  tbl.filter(required_columns);

  cout << tbl.get_row_count() << " valid rows." << endl;

  tbl.get_col(namecol, namevec);
  tbl.get_col(valcol,  weightvec_str);
  tbl.get_col(seqcol,  seqvec);
  tbl.get_col(seqcol2,  seqvec2);

  if(!regex.empty()){
    transform(seqvec.begin(), seqvec.end(), seqvec.begin(), mysubstr(regex));
  }
  if(!regex2.empty()){
    transform(seqvec2.begin(), seqvec2.end(), seqvec2.begin(), mysubstr(regex2));
  }

  /* set strings of weightvec_str to double weightvec */
  weightvec.resize(weightvec_str.size());
  transform(weightvec_str.begin(), weightvec_str.end(),
	    weightvec.begin(), mystrtod());

  double variance_correction;
  variance_correction = Utils::to_ranks_of(weightvec, rankvec);
  cout << "Variance Correction: " << variance_correction << endl;
  for(i = 0; i < tbl.get_row_count(); i++){
    cout << i << "\t" << rankvec[i] << "\t" << namevec[i] << "\t" << weightvec[i] << "\t" << seqvec[i] << endl;
  }
 
  /********************* END  : Read and process data file *********************/

  /********************* BEGIN: creating suffix array *********************/
  cout << "Creating suffix array 1 ..." << flush;
  boost::timer t2;
  SuffixArray sa(seqvec);
  if(maxpatl > 0){
    cout << "... trimming lcp ..." << flush;
    sa.trim(maxpatl);
  }
  cout << "Done: " << t2.elapsed() << " seconds." << endl;


  cout << "Creating suffix array 2 ..." << flush;
  SuffixArray sa2(seqvec2);
  if(maxpatl > 0){
    cout << "... trimming lcp ..." << flush;
    sa2.trim(maxpatl);
  }
  cout << "Done: " << t2.elapsed() << " seconds." << endl;
  /********************* END  : creating suffix array *********************/

  if(combo){ 
    /******************************* pattern pairs *******************************/
    vector<SAColorSetSizeResultPair<TaggedNumericDataValue> > results_pair;
    SuffixArrayIterator sa_it(sa);
    SASubstring tmppat;
    vector<SASubstring> pattern1_candidates;
    tmppat = sa_it.get_next(false);
    while(tmppat.l || tmppat.h){
      pattern1_candidates.push_back(tmppat);
      tmppat = sa_it.get_next(false);
    }

    if(use_naive){
      vector<vector<bool> > matchresults(pattern1_candidates.size());
      boost::timer t;
      cout << "Allocating memory" << endl;
      for(i = 0; i < matchresults.size(); i++){
	matchresults[i].resize(seqvec.size());
      }
      cout << "Calculating matches for all patterns" << endl;
      for(i = 0; i < matchresults.size(); i++){
	cout << "Pattern: " << i << "/" << pattern1_candidates.size() << endl;
	for(j = 0; j < seqvec.size(); j++){
	  matchresults[i][j] = (seqvec[j].find(sa.get_string().c_str() + sa.get_suffix(pattern1_candidates[i].l),
					       0, pattern1_candidates[i].h) != std::string::npos);
	}
      }
      cout << t.elapsed() << " seconds for all match results" << endl;
      
      size_t hogecnt;
      double sum;
      double maxscore = 0.0;
      for(i = 0; i < matchresults.size(); i++){
	cout << "Match: " << i << "/" << pattern1_candidates.size() << endl;
	for(j = 0; j < matchresults.size(); j++){
	  sum = 0.0;
	  hogecnt = 0;
	  for(size_t k = 0; k < seqvec.size(); k++){
	    if(matchresults[i][k] || matchresults[j][k]){
	      hogecnt++;
	      sum += rankvec[k];
	    }
	  }
	  double s = Scores::ranksumtest(hogecnt, seqvec.size()-hogecnt, sum, variance_correction);
	  if (s > maxscore) maxscore = s;
	}
      }
      cout << "Max score: " << maxscore << endl;
      cout << t.elapsed() << " seconds for all pairs (virtual) calculation" << endl;
    } else {
      pthread_attr_t thread_attr;
      vector<pthread_t> threads;
      vector<struct thread_args> thread_args;
      size_t jobsizebase, jobsizerem, jobid = 0;
      concurrency = min(concurrency, pattern1_candidates.size());
      jobsizebase = pattern1_candidates.size() / concurrency;
      jobsizerem  = pattern1_candidates.size() % concurrency;
      threads.resize(concurrency);
      thread_args.resize(concurrency);
      pthread_attr_init(&thread_attr);
      pthread_attr_setscope(&thread_attr, PTHREAD_SCOPE_SYSTEM);
      for(i = 0; i < concurrency; i++){
	cout << "prepare launch for thread: " << i << endl;
	thread_args[i].id = i;
	thread_args[i].job_begin = jobid;
	thread_args[i].job_end = jobid + jobsizebase + ((i < jobsizerem) ? 1 : 0);
	thread_args[i].result_num = result_num;
	thread_args[i].combo = combo;
	
	thread_args[i].sa      = &sa;
	thread_args[i].sa2      = &sa2;
	thread_args[i].p1vec   = &pattern1_candidates;
	thread_args[i].seqvec  = &seqvec;
	thread_args[i].rankvec = &rankvec;
	
	thread_args[i].vc = variance_correction;
	cout << "calling pthread_create for thread: " << i << endl;
	pthread_create(&(threads[i]), &thread_attr, calc_thread, &(thread_args[i]));
	cout << "done calling pthread_create for thread: " << i << endl;
	jobid = thread_args[i].job_end;
      }
      
      for(i = 0; i < concurrency; i++){
	vector<SAColorSetSizeResultPair<TaggedNumericDataValue> > * tmpresvec;
	SAColorSetSizeResultPair<TaggedNumericDataValue> tmpresval;
	cout << "calling pthread_join for thread: " << i << endl;
	pthread_join(threads[i], (void **)&tmpresvec);
	cout << "done calling pthread_join for thread: " << i << endl;
	for(j = 0; j < tmpresvec->size(); j++){
	  if(results_pair.size() < result_num){
	    results_pair.push_back((*tmpresvec)[j]);
	    push_heap(results_pair.begin(), results_pair.end(), cmppair);
	  } else if(results_pair.front().score < (*tmpresvec)[j].score){
	    pop_heap(results_pair.begin(), results_pair.end(), cmppair);
	    results_pair.back().score = (*tmpresvec)[j].score;
	    results_pair.back().p1    = (*tmpresvec)[j].p1;
	    results_pair.back().p2    = (*tmpresvec)[j].p2;
	    results_pair.back().dv    = (*tmpresvec)[j].dv;
	    results_pair.back().total = (*tmpresvec)[j].total;
	    push_heap(results_pair.begin(), results_pair.end(), cmppair);
	  }
	}
	delete(tmpresvec);
      }
    }
    
    sort(results_pair.begin(), results_pair.end(), cmppair);

    // output results ////////////////////
    const TNDVRankSumTestScore ranksumtest(combo);
    for(i = 0; i < results_pair.size(); i++){
      cout << i << ": score: " << results_pair[i].score
	   << " (" << results_pair[i].p1.l
	   << "," << results_pair[i].p1.h
	   << ") "
	   << "'" << sa.get_substr(results_pair[i].p1) << "'"
	   << " op "
	   << " (" << results_pair[i].p2.l
	   << "," << results_pair[i].p2.h
	   << ") "
	   << "'" << sa.get_substr(results_pair[i].p2) << "'"
	   << " [" << ranksumtest.eval_str(results_pair[i].dv, results_pair[i].total) << "]" 
	   << endl;
    }
  } else {
    /******************************* single pattern *******************************/
    boost::timer t;
    NumericDataValueFactory fctry(rankvec);
    vector<SAColorSetSizeResult<NumericDataValue> > results;
    NDVRankSumTestScore ranksumtest;
    results = SAColorSetSize<NumericDataValue,
                             NumericDataValueFactory,
                             NDVRankSumTestScore>::solve(sa, fctry, ranksumtest, result_num, false);
    // output results ////////////////////
    for(i = 0; i < results.size(); i++){
      cout << i << ": score: " << results[i].score
	   << " (" << results[i].node.l
	   << "," << results[i].node.h
	   << ") "
	   << "'" << sa.get_substr(results[i].node) << "'"
	   << " [" << ranksumtest.eval_str(results[i].dv, fctry.get_total()) << "]" 
	   << endl;
    }
    cout << t.elapsed() << " secs." << endl;
  }
  return 0;
}
