/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "DBTable.h"
#include "StringUtils.h"
#include <iostream>
#include <fstream>
#include <cassert>

LineProcessor::LineProcessor(char dc, char cc)
: delimiter_chars(1,dc), comment_chars(1,cc){
}

LineProcessor::LineProcessor(const std::string & dc, const std::string & cc)
: delimiter_chars(dc), comment_chars(cc){
}

LineProcessor::~LineProcessor(){
}

std::vector<std::string> LineProcessor::process_line(std::string & line) const {
  std::vector<std::string> result;
  if(line.empty() || comment_chars.find(line[0]) != std::string::npos)
    return result;
  return (StringUtils::split(line, delimiter_chars));
}

DBTable::DBTable(const std::string & filename,
		 const std::string & delimiter){
  std::string line;
  std::ifstream f(filename.c_str(), std::fstream::in);
  if(!f.good()){
    std::cerr << "Error: DBTable::DBTable(): Failed to open file: '"
	      << filename << "'" << std::endl;
  }
  do {
    std::getline(f, line);
    this->table.push_back(StringUtils::split(line, delimiter));
  } while (!f.eof());
  return;
}

DBTable::DBTable(const std::string & filename,
		 const LineProcessor & processor){
  std::string line;
  std::ifstream f(filename.c_str(), std::fstream::in);
  if(!f.good()){
    std::cerr << "Error: DBTable::DBTable(): Failed to open file: '"
	      << filename << "'" << std::endl;
  } else {
    do {
      std::getline(f, line);
      std::vector<std::string> newrow = processor.process_line(line);
      if(!newrow.empty()) this->table.push_back(newrow);
    } while (!f.eof());
  }
  return;
}

DBTable::~DBTable()
{}

std::size_t DBTable::get_row_count() const {
  return this->table.size();
}

const std::vector<std::string> & DBTable::get_row(std::size_t row) const {
  assert(row < get_row_count());
  return this->table[row];
}

void DBTable::get_row(std::size_t row, std::vector<std::string> & v) const {
  assert(row < get_row_count());
  v = get_row(row);
}

void DBTable::get_col(std::size_t col, std::vector<std::string> & v) const {
  std::size_t i;
  v.clear();
  for(i = 0; i < get_row_count(); i++){
    const std::vector<std::string> & row = get_row(i);
    if(col < row.size()) v.push_back(row[col]);
    else v.push_back("");
  }
}

void DBTable::del_row(std::size_t row){
  this->table.erase(this->table.begin() + row);
}

void DBTable::filter(const std::vector<std::size_t> & required_columns){
  std::vector<std::vector<std::string> > newtable;
  std::size_t i, colp;
  for(i = 0; i < get_row_count(); i++){
    const std::vector<std::string> & row = get_row(i);
    for(colp = 0; colp < required_columns.size(); colp++){
      if(required_columns[colp] >= row.size() || row[required_columns[colp]].empty())
	break;
    }
    if(colp == required_columns.size()){
      newtable.push_back(row);
    }
  }
  this->table = newtable;
}

void DBTable::get(std::size_t row, std::size_t col, std::string & s) const {
  assert(row < get_row_count());
  const std::vector<std::string> & r = get_row(row);  
  if(col < r.size()) s = r[col];
  else s = "";
  return;
}

void DBTable::set(std::size_t row, std::size_t col, const std::string & s){
  assert(row < get_row_count());
  std::vector<std::string> & r = table[row];
  if(col >= r.size())
    r.resize(col+1);
  r[col] = s;
  return;
}
