/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#ifndef DBTABLE_H
#define DBTABLE_H

#include <string>
#include <vector>
#include <cstddef>

/** class for processing lines. 
 * The base class just splits lines into elements according to delimiters
 * or ignores lines beginning with comment characters. */
class LineProcessor {
  std::string delimiter_chars;
  std::string comment_chars;
 public:
  LineProcessor(char dc = '\t', char cc = '#');
  LineProcessor(const std::string & dc = "\t", const std::string & cc = "#");
  
  /** Processes a single line to extract elements out of it.*/
  virtual std::vector<std::string> process_line(std::string & line) const;
  virtual ~LineProcessor();
};

/** Class for row-oriented data tables. 
    Each row can have different column counts. */
class DBTable {
  std::vector<std::vector<std::string> > table;
public:
  /** Create new table based on file. Each line of the file
   * defines a single row, where each element is separated by delimiter. */
  DBTable(const std::string & filename, 
	  const std::string & delimiter = "\t");
  
  /** create a new table based on file. Each line of the file
   * is processed by processor.process_line(), which returns
   * std::vector<std::string>, a vector of elements. The line is
   * ignored if the vector is empty. */
  DBTable(const std::string & filename,
	  const LineProcessor & processor);
  
  /** return the number of rows in the table */
  std::size_t get_row_count() const;			

  /** return reference to rowth row of the table */
  const std::vector<std::string> & get_row(std::size_t row) const;

  /** set values of rowth row to v */
  void get_row(std::size_t row, std::vector<std::string> & v) const;

  /** set values of colth column to v.
      an empty string is set for rows not containing enough columns */
  void get_col(std::size_t col, std::vector<std::string> & v) const;

  /** remove row from table */
  void del_row(std::size_t row);

  void get(std::size_t row, std::size_t col, std::string & s) const;

  void set(std::size_t row, std::size_t col, const std::string & s);

  /** delete rows which do not contain all elements for the following columns */
  void filter(const std::vector<std::size_t> & required_columns);

  virtual ~DBTable();
};

template<typename funcTp>
void DBTable_apply_column(DBTable & tbl, std::size_t col, const funcTp & f){
  std::size_t row;
  for(row = 0; row < tbl.get_row_count(); row++){
    std::string s;
    tbl.get(row, col, s);
    f(s);
    tbl.set(row, col, s);
  }
}
#endif // DBTABLE_H
