/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "RangeMinimumTable.h"
#include <stack>
#include <cassert>
#include <iostream>

static unsigned int clearbits[32] = {
  0xffffffff, 0xfffffffe, 0xfffffffc, 0xfffffff8,
  0xfffffff0, 0xffffffe0, 0xffffffc0, 0xffffff80, 
  0xffffff00, 0xfffffe00, 0xfffffc00, 0xfffff800,
  0xfffff000, 0xffffe000, 0xffffc000, 0xffff8000, 
  0xffff0000, 0xfffe0000, 0xfffc0000, 0xfff80000,
  0xfff00000, 0xffe00000, 0xffc00000, 0xff800000, 
  0xff000000, 0xfe000000, 0xfc000000, 0xf8000000,
  0xf0000000, 0xe0000000, 0xc0000000, 0x80000000
};

/* return floor of log n.
   (This should be implemented with a Bit Scan Reverse on Intel platforms) */
static inline int intlog2(int n){
  int res;
  assert(n > 0);
#ifdef INTEL
  asm volatile ("bsr %1,%0\n\t" : "=r"(res) : "r"(n));
  return(res);
#else
  for(res = 0; n > 0; n >>= 1, res++);
  return(res-1);
#endif
}

/* return position of least significant set bit.
   will loop infinitely if n is 0. 
   (This shold be implemented with a Bit Scan Foward on Intel Platforms) */
static inline int lsbset (int n){
  int res = 0;
#ifdef INTEL
  asm volatile ("bsf %1,%0\n\t" : "=r"(res) : "r"(n));
#else
  while((n & 0x1) == 0){
    res++; n >>= 1;
  }
#endif
  return(res);
}

RangeMinimumTable::RangeMinimumTable(const std::vector<std::size_t> & orig_array){
	std::size_t i, j, rows, cols, block_cnt, tmp, size;
	std::vector<std::size_t> block_minpos;  /*  32 ̃ubNꂼ̍ŏn̈ʒu */
	std::vector<std::size_t> g;
	std::stack<int> gstack;

	size = orig_array.size();
	assert(size >= 1);
	assert(sizeof(int) >= 4);

	array = orig_array;

	/* divide input array into blocks of size 32.
       block_cnt is the number of such blocks.
       block_minpos is an array that contains the 
       minimum positions in each block. */
	block_cnt = ((size-1) >> 5) + 1;
	block_min = std::vector<std::size_t>(block_cnt);
	for(i = j = 0; i < size; i++){
    	if(i % 32 == 0){
			if(i > 0) j++;
			block_min[j] = i;
		} else if(array[i] < array[block_min[j]]) {
      		block_min[j] = i;
		}
	}
	
  	/* make a sparse table for the rmq of the blocks.
      sparse[j][i] represents the minimum in
      block[i] to block[i + 2^{j+1} - 1] */
  	rows = intlog2(block_cnt);
  	sparse = std::vector<std::vector<std::size_t> >(rows);
  	
	/* sparse tables aren't neede when the 
	 * array is less than 32 elements long */
	if(rows > 0){
		/* first row is min of adjacent entries.
       	   table entries are converted to positions in original array */
  		sparse[0] = std::vector<std::size_t>(block_cnt - 1);
	  	for(i = 0; i < block_cnt - 1; i++){
    		if(array[block_min[i]] < array[block_min[i+1]])
      			sparse[0][i] = block_min[i];
	    	else
	      		sparse[0][i] = block_min[i+1];
	  	}
	  	for(j = 1; j < rows; j++){
    		int rowelmlen = 2 << j;    /* 2^{j+1} */
	    	cols = block_cnt - rowelmlen + 1;
	    	sparse[j] = std::vector<std::size_t>(cols);
	    	for(i = 0; i < cols; i++){
	      		if(array[sparse[j-1][i]] < array[sparse[j-1][i + (rowelmlen >> 1)]])
					sparse[j][i] = sparse[j-1][i];
	      		else
					sparse[j][i] = sparse[j-1][i + (rowelmlen >> 1)];
	    	}
	  	}
	}

  	/* create integers for constant time rmq inside the blocks
       In each block:
       - g[i]: the first position to the left of i
       where array[g[i]] < array[i] (or -1 if there is no such position).
       - l[i]: the jth bit of l[i] is 1 iff j is the first
       position left of i where array[j] < array[i] */

	g = std::vector<std::size_t>(size);
  	labels = std::vector<std::size_t>(size);
  	for(i = j = 0; i < size; i++){
    	if(i % 32 == 0){ 
      		g[i] = 0xffffffff;   /* first g in a block is always -1 */
      		labels[i] = 0;       /* first label in a block is always 0  */
      		gstack.push(i);
    	} else {
      		if(array[i] > array[i-1]){
				g[i] = i-1;
				gstack.push(i);
      		} else {
				g[i] = 0xffffffff;
				while(gstack.size() > 0){
	  				tmp = gstack.top();
				  	if(array[tmp] < array[i]){
	    				g[i] = tmp; break;
	  				}
	  				gstack.pop();
				}
				gstack.push(i);
      		}
      		if(g[i] == 0xffffffff) labels[i] = 0;
      		else labels[i] = (labels[g[i]] | (1 << g[i]));
    	}
    	if(i % 32 == 31){
      		while(gstack.size() > 0) gstack.pop();
      		j++;
    	}
  	}
  	return;
}

std::size_t RangeMinimumTable::query(std::size_t l, std::size_t r) const {
  std::size_t blocknum_l, blocknum_r, blockdiff, blockmin;
  std::size_t x, v, pos;
  std::size_t v1, v2, pos1, pos2;
  if(l == r) return l;
  assert(l < r);  
  blocknum_l = (l >> 5); blocknum_r = (r >> 5);  /* obtain which blocks l and r will come in */
  switch(blockdiff = blocknum_r - blocknum_l){
  case 0: /* one inblock query in block blocknum_l. from position (l%32) to (r%32) */
    x = l % 32;
    v = labels[r] & clearbits[x] ; /* clear (x - 1) insiginificant bits */
    if(v == 0) return(r);
    /* position of least significant bit is answer */
    pos = lsbset(v);
    return ((blocknum_l << 5) + pos);
    break;
  case 1: /* two inblock queries. in block blocknum_l and blocknum_r.
	     in blocknum_l: postion (l%32) to 31
	     in blocknum_r: postion 0 to (r%32) */
    v1 = labels[(blocknum_l<<5)+31] & clearbits[l%32];
    v2 = labels[r];
    pos1 = (v1 == 0) ? (blocknum_l<<5)+31 : ((blocknum_l<<5) + lsbset(v1));
    pos2 = (v2 == 0) ? r : lsbset(v2) + (blocknum_r<<5);
    return((array[pos1] < array[pos2]) ? pos1 : pos2);
    break;
  default: /* two inblock queries, and a query over the blocks.
	      in blocknum_l: postion (l%32) to 31
	      in blocknum_r: postion 0 to (r%32).
	      block (blocknum_l+1) to (blonum_r-1) */
    v1 = labels[(blocknum_l<<5)+31] & clearbits[l%32];
    v2 = labels[r];
    pos1 = (v1 == 0) ? (blocknum_l<<5)+31 : ((blocknum_l<<5) + lsbset(v1));
    pos2 = (v2 == 0) ? r : lsbset(v2) + (blocknum_r<<5);
    if(blockdiff == 2){ /* (blocknum_l+1) == (blonum_r-1) */
      blockmin = block_min[blocknum_l+1];
    } else {
      /* rmq of blocknum_l+1 to blocknum_r-1 */
      int t1, t2, k;
      k = intlog2(blockdiff-1) - 1; assert(k >= 0);
      t1 = sparse[k][blocknum_l+1];
      t2 = sparse[k][blocknum_r - (1 << (k+1))];
      blockmin = (array[t1] < array[t2]) ? t1 : t2;
    }
    pos1 = (array[pos1] < array[pos2]) ? pos1 : pos2;
    return((array[pos1] < array[blockmin]) ? pos1 : blockmin);
  }
}
