/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#ifndef SACOLORSETSIZE_H
#define SACOLORSETSIZE_H
#include <limits>
#include "SuffixArray.h"

/** A template structure that holds information about
 * the results of a color set size problem, solved by
 * suffix arrays. */
template <class DataValue>
struct SAColorSetSizeResult {
  double score;
  SASubstring node;
  DataValue dv;
};

/** Solves the Generalized Color Set Size Problem using Suffix Array.
 * DataValue is the generalized value of the color set
 * DataValue should implment the folowing constructor/methods:
 * - DataValue() constructor (not really used).
 * - void accumulate(const DataValue & v);
 * - void subtract(const DataValue& v);
 * - static score(const DataValue & match)
 * - static DataValue zero
 * 
 * DataValueFactory should implement the following methods:
 * - create_datavalue(std::size_t id);
 */
template <class DataValue, class DataValueFactory, class Fitness>
class SAColorSetSize {
  /// private structure to handle data values and node
  struct SADataNode { SASubstring node; DataValue dv; };
  /// private structure to compare scored DataValues
  struct CmpPair 
    : public std::binary_function<SAColorSetSizeResult<DataValue>,
    SAColorSetSizeResult<DataValue>, bool> {
    bool operator()(const SAColorSetSizeResult<DataValue> & x,
                    const SAColorSetSizeResult<DataValue> & y){
      return ((x.score > y.score) ? true : false);
    }
  };
  /// calculate and set correction factors using rmq
  static void calculate_correction_factors(
    const SuffixArray & sa, std::vector<DataValue> & cf,
    const DataValueFactory & dvf
  );
public:
  /** Solve the generalized Color Set Size Problem using suffix array sa
   * and DataValueFactory dvf.
   * \param sa The Generalized Suffix array.
   * \param dvf A class which createse DataValue instances: zero, total, and for each id.
   * \param keepsize How many top scoring nodes to keep.
   * \param report_leaf Whether or not to report leaves.
   */
  static std::vector<SAColorSetSizeResult<DataValue> > 
    solve(const SuffixArray & sa,
	  const DataValueFactory & dvf,
	  const Fitness & scoref,
	  std::vector<DataValue> & correction_factors,	  
	  const std::size_t keepsize, const bool report_leaf = true);

  static std::vector<SAColorSetSizeResult<DataValue> > 
    solve(const SuffixArray & sa,
	  const DataValueFactory & dvf,
	  const Fitness & scoref,
	  const std::size_t keepsize, const bool report_leaf = true){
    std::vector<DataValue> correction_factors;
    return (SAColorSetSize::solve(sa, dvf, scoref, correction_factors, keepsize, report_leaf));
  }
};



////////////////////////////////////////////////////////////////////
// Implementation
////////////////////////////////////////////////////////////////////

/*********** BEGIN: SAColorSetSize::calculate_correction_factors ***********/
template <class DataValue, class DataValueFactory, class Fitness>
  void SAColorSetSize<DataValue,DataValueFactory,Fitness>::calculate_correction_factors(
    const SuffixArray & sa, std::vector<DataValue> & cf,
    const DataValueFactory & dvf
  ) {
  std::fill(cf.begin(), cf.end(), DataValue::zero);
  std::vector<std::size_t> prevpos(sa.get_strcnt(), std::numeric_limits<std::size_t>::max());
  for(std::size_t i = 0; i < sa.size(); i++){
    std::size_t id = sa.get_id(i);
    if(prevpos[id] < i){
      std::size_t pos = sa.rmq(prevpos[id]+1, i);
      cf[pos].accumulate(dvf.create_datavalue(id));
    }
    prevpos[id] = i;
  }
}

/*********** END  : SAColorSetSize::calculate_correction_factors ***********/

/*********** BEGIN: SAColorSetSize::solve ***********/
template <class DataValue, class DataValueFactory, class Fitness>
  std::vector<SAColorSetSizeResult<DataValue> > 
SAColorSetSize<DataValue,DataValueFactory,Fitness>::solve(const SuffixArray & sa,
							  const DataValueFactory & dvf,
							  const Fitness & scoref,
							  std::vector<DataValue> & correction_factors,
							  const std::size_t keepsize, const bool report_leaf){
  std::vector<SAColorSetSizeResult<DataValue> > result;
  std::vector<SADataNode> stack;
  std::size_t index;
  SADataNode new_node, report_node;
  SAColorSetSizeResult<DataValue> resval;
  DataValue dvsum;
  DataValue total_val = dvf.get_total();

  assert(keepsize > 0);
  assert(sa.size() > 1);
  
  correction_factors.resize(sa.size());

  /*********** BEGIN: initializations ***********/
  /* reserve memory space for result */
  result.reserve(keepsize);
  /* push first leaf onto stack */    
  new_node.node.l = 0; new_node.node.h = sa.size() - sa.get_suffix(0);
  new_node.dv = dvf.create_datavalue(sa.get_id(0));
  stack.push_back(new_node);
  /* calculate correction factors */
  calculate_correction_factors(sa, correction_factors, dvf);
  /*********** END  : initializations ***********/
  
  /*********** BEGIN: Main Loop ***********/
  for(index = 1; index <= sa.size(); index++){
    dvsum = DataValue::zero;
    /* Initialize new_node: lca of leaf i-1 and leaf i */
    new_node.node.l = index - 1;
    if(index < sa.size()) new_node.node.h = sa.get_lcp(index);
    
    /* Pop and report nodes in the stack if they are deeper than the new node.
     * index == sa.size() indicates that this is the last loop, so we
     * should pop and report all the nodes in the stack. */
    while(!stack.empty()){
      report_node = stack.back();
      /* exit loop if deepest node (on the stack top) is shallower than newnode */
      if((report_node.node.h <= new_node.node.h) && (index < sa.size())) break;
      /* update new_node (left position and DataValue accumulation ) */
      new_node.node.l = report_node.node.l;
      dvsum.accumulate(report_node.dv);
      /* reflect accumulation to report node */
      report_node.dv = dvsum;
      if(report_leaf /* report (save) report_node if wanted */
	 || (report_node.node.h != sa.size() - sa.get_suffix(report_node.node.l)
	     && report_node.node.h != 0)){
	double score = scoref.score(dvsum,total_val);
	if(result.size() < keepsize){
	  resval.score = score;
	  resval.dv = report_node.dv;
	  resval.node = report_node.node;
	  result.push_back(resval);
	  std::push_heap(result.begin(), result.end(), CmpPair());
	} else if(result.front().score < score) {
	  std::pop_heap(result.begin(), result.end(), CmpPair());
	  result.back().score = score;
	  result.back().node  = report_node.node;
	  result.back().dv    = report_node.dv;
	  std::push_heap(result.begin(), result.end(), CmpPair());
	}
      }
      stack.pop_back(); /* discard (reported) top element */
    }
    if(index < sa.size()){
      dvsum.subtract(correction_factors[index]); // subtract correction factors at this node
      if(stack.empty() || stack.back().node.h < new_node.node.h){
	// a completely new node
	new_node.dv = dvsum;
	stack.push_back(new_node);
      } else {
	// or accumulate values to current stack top.
	stack.back().dv.accumulate(dvsum);
      }
      new_node.node.l = index; new_node.node.h = sa.size() - sa.get_suffix(index);
      new_node.dv = dvf.create_datavalue(sa.get_id(index));
      stack.push_back(new_node); // push leaf
    }
  }
  /*********** END  : Main Loop ***********/
  
  // resort result by score (not heap)
  std::sort(result.begin(), result.end(), CmpPair());
  return result;
}
/*********** END  : SAColorSetSize::solve ***********/

#endif //SACOLORSETSIZE_H
