/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "SuffixArray.h"
#include <algorithm>
#include <vector>
#include <functional>
#include <cstring>
#include <iostream>

/** a compare function for the suffixes of strings */
struct cmpsubstr : public std::binary_function<char *, char*, bool> {
  bool operator()(const char * sp1, const char * sp2) {
    int c = strcmp(sp1, sp2);
    if(c == 0){
      return(sp1 > sp2);
    } else {
      return(c < 0);
    }
  }
};

void SuffixArray::construct(){
  std::size_t i, h, x, j, id; 
  std::size_t strlen;
  const char * text;

  strlen = orig_string.length (); /* '\0' should be included */
  text = orig_string.c_str();     /* null terminated pointer to string */

  sa    = std::vector<std::size_t>(strlen);
  lcpa  = std::vector<std::size_t>(strlen);
  ranka = std::vector<std::size_t>(strlen);
  ida   = std::vector<std::size_t>(strlen);

  std::vector<const char *> tmp(strlen);
  /* sort the suffixes using a very naive algorithm */
  for(i = 0; i < strlen; i++) tmp[i] = text + i;
  std::sort(tmp.begin(), tmp.end(), cmpsubstr());
  for(i = 0; i < strlen; i++) sa[i] = tmp[i] - text;

  /* compute rank array */
  for(i = 0; i < strlen; i++) ranka[sa[i]] = i;

  /* compute lcp array */
  for(h = i = 0; i < strlen; i++){
    x = ranka[i];
    if(x > 0){
      const char * p0, * p1;    
      j = sa[x-1];
      p1 = text + i + h;
      p0 = text + j + h;
      while((*p1 == *p0) && (*p0 != '\0')){
        p1++; p0++; h++;
      }
      lcpa[x] = h;
      if(h > 0) h--;
    }
  }
  
  /* assign id */
  for(id = i = 0; i < strlen; i++){
    ida[ranka[i]] = id;
    if(text[i] == '\0') id++;
  }
  strcnt = id;
  
  /* construct RangeMinimumTable for lcpa */
  rmt = new RangeMinimumTable(lcpa);

  return;
}

SuffixArray::SuffixArray(const std::string & str){
  orig_string = str;
  if(orig_string[orig_string.length()-1] != '\0')
    orig_string.push_back('\0');
  SuffixArray::construct();
}

SuffixArray::SuffixArray(const std::vector<std::string> & strvec){
  std::vector<std::string>::const_iterator it;
  for(it = strvec.begin(); it != strvec.end(); it++){
    orig_string.append(*it);
    if(orig_string[orig_string.length()-1] != '\0')
      orig_string.push_back('\0');
  }
  SuffixArray::construct();
}

SuffixArray::~SuffixArray(){
  delete this->rmt; 
}

void SuffixArray::trim(std::size_t k) {
  std::vector<std::size_t>::iterator it;
  if(k > 0)
    for(it = lcpa.begin(); it != lcpa.end(); it++)
      if((*it) > k) *it = k;
}


std::size_t SuffixArray::rmq(std::size_t i, std::size_t j) const {
  return rmt->query(i, j);
}

SuffixArrayIterator::SuffixArrayIterator(const SuffixArray & a): sa(a), index(1) {
  SASubstring firstleaf;
  assert(sa.size() > 1);
  /* insert first leaf into stack */
  firstleaf.l = 0; firstleaf.h = this->sa.size() - this->sa.get_suffix(0);
  this->stack.push_back(firstleaf);
  /* the first new_node is the lca between the first two nodes */
  new_node.l = 0;
  new_node.h = this->sa.get_lcp(1);
}

struct SASubstring SuffixArrayIterator::get_next(bool report_leaf){
  struct SASubstring report_node;
  while(index <= this->sa.size()){
    /* IF
     * - the stack is not empty, AND
     * -- the current new_node is shallower than the top stack element,
     * -- OR index >= this->sa.size() (we should pop all elements) 
     * THEN
     *   Report the top stack element. */
    while(!this->stack.empty()){
      report_node = this->stack.back();
      if((report_node.h > new_node.h)
	 || (index >= this->sa.size())){
	new_node.l = report_node.l;
	this->stack.pop_back();
	if(report_leaf ||
	   (report_node.h != sa.size() - sa.get_suffix(report_node.l)
	    && report_node.h != 0))
	  return(report_node);
      } else {
	break;
      }
    }
    /* else, we should add the new node and leaf,
     *  corresponding to the index, to the stack */
    if(index < this->sa.size()){
      // only add if the newnode is strictly deeper than the top.
      // equal depth means that they are the same node.
      if(this->stack.empty() || this->stack.back().h < new_node.h){
	this->stack.push_back(new_node);
      }
      // add the leaf to the stack.
      new_node.l = index; new_node.h = this->sa.size() - this->sa.get_suffix(index);
      this->stack.push_back(new_node);
    }
    /* finally, increment the index, and update new_node to be
     * the lca for the next consecutive leaves */
    index++;
    new_node.l  = index - 1;
    if(index < this->sa.size())
      new_node.h = this->sa.get_lcp(index);
  }
  report_node.l = report_node.h = 0;
  return(report_node);
}

// void SuffixArrayIterator::for_all(
//  	for(index = 1; index <= this->sa.size(); index++){
//  		std::cout << "index: " << index << std::endl;
//		/* new_node = lca of leaf i-1 and leaf i */
//		new_node.first  = index-1;
//		if(index < this->sa.size())
//			new_node.second = this->sa.get_lcp(index);
//
//		/* Vm[h[z͂ǂǂ pop BA
//		 *  index == this->sa.size() ̎́AS pop  */
//		while(!this->stack.empty()){
//       		std::pair<std::size_t, std::size_t> report_node = this->stack.back();
//       		if((report_node.second <= new_node.second)
//		       && index < this->sa.size())
//		    	break;
//
//    		new_node.first = report_node.first; /* update left for new node*/
//    		/* report the pattern */
//			std::cout << "(" << report_node.first
//			          << "," << report_node.second << ")" << std::endl;
//			this->stack.pop_back();                   /* remove top element */
//		}
//		if(index < this->sa.size()){
//			/* a new node between old nodes */
//			if(this->stack.empty() || this->stack.back().second < new_node.second){
//				this->stack.push_back(new_node);
//			}
//			
//			/* push leaf. */
//			this->stack.push_back(
//				std::pair<std::size_t,std::size_t>(
//					index, 
//					this->sa.size() - this->sa.get_suffix(index)
//				)
//			);
//		}
//  	}
//	std::cout << "Finished" << std::endl;
// 	return(new_node);
//}
