/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#ifndef SUFFIXARRAY_H
#define SUFFIXARRAY_H
#include <string>
#include <vector>
#include <cassert>
#include <iostream>
#include "RangeMinimumTable.h"

/** a struct that represents a substring of the text
 * with respect to the suffix array. */
struct SASubstring {
  std::size_t l; ///< position in suffix array
  std::size_t h; ///< length of path
};

/** A class for generalized suffix arrays.
 */
class SuffixArray {
  std::string orig_string;
  
  std::vector<std::size_t> sa;      ///< position -> suffix array
  std::vector<std::size_t> lcpa;    ///< position -> lcp  array
  std::vector<std::size_t> ranka;   ///< suffix   -> rank array
  std::vector<std::size_t> ida;     ///< position -> id   array
  
  RangeMinimumTable * rmt;          ///< range minimum table for lcp array
  
  std::size_t	strcnt;               ///< # of strings
  
  // construct suffix array for this->orig_string.
  // Assume that orig_string is '\0' terminated.
  // orig_string.length() should return the length including the last '\0'.
  // Multiple strings should also be separated by '\0's.
  void construct();
public:
  /** Construct the suffix array for a given string */
  SuffixArray(const std::string & str);
  
  /** Construct the suffix array for a vector of strings */
  SuffixArray(const std::vector<std::string> & strvec);
  
  /// return size of arrays (string)
  inline std::size_t size() const {
    return (sa.size());
  }
  
  /// return the suffix at a given rank (element of suffix array at given position)
  inline std::size_t get_suffix(std::size_t pos) const {
    assert(pos < this->size());
    return (sa[pos]);
  }
  
  /// return the rank (position in suffix array) for a given suffix
  inline std::size_t get_rank(std::size_t suffix) const {
    assert(suffix < this->size());
    return (ranka[suffix]);
  }
  
  /// return the lcp for a position in the suffix array
  inline std::size_t get_lcp (std::size_t pos) const {
    assert(pos < this->size());
    return (lcpa[pos]);
  }
  
  /// return the id of the given suffix.
  inline std::size_t get_id(std::size_t pos) const {
    assert(pos < this->size());
    return (ida[pos]);
  }
  
  /// return the number of strings
  inline std::size_t get_strcnt() const {
    return strcnt;
  }

  /** `trim' the lcp array to maximum value k. 
      rmq need not change. */
  void trim(std::size_t k);
  
  /// Return the position k such that lcp[k] = min\{ lcp[k'] : i \leq k'\leq j\}.
  std::size_t rmq(std::size_t i, std::size_t j) const;
  
  /// Return reference to original string
  const std::string & get_string() const { return orig_string; }

  const std::string get_substr(struct SASubstring s) const {
    return orig_string.substr(get_suffix(s.l), s.h);
  }
  
  /// Destructor. Get rid of range minimum table.
  ~SuffixArray();
};

/** Iterator for bottom-up traveral of suffix tree using suffix array. */
class SuffixArrayIterator {
  const SuffixArray & sa;
  std::size_t index;
  std::vector<struct SASubstring> stack;
  struct SASubstring new_node;
public:
  SuffixArrayIterator(const SuffixArray & a);
  
  /** Returns next position in a post-order traversal of the suffix tree.
   * The last element is always the root, expressed by:
   * std::pair<std::size_t,std::size_t>(0,0); */
  struct SASubstring get_next(bool report_leaf = true);
  std::size_t get_index() const { return index; };
};

template <typename Func>
void iter(const SuffixArray & a, const Func & f, bool report_leaf = true){
  SuffixArrayIterator it(a);
  SASubstring n;
  while(n=it.get_next(), (n.l||n.h)){
    f(n);
  }
  return;
}

template <typename Func,class DataValue>
void fold(const Func & f, DataValue & accum, bool report_leaf = true){
  SuffixArrayIterator it(a);
  SASubstring n;
  while(n=it.get_next(), (n.l||n.h)){
    accum = f(n, accum);
  }
  return;
}
#endif // SUFFIXARRAY_H
