/*
  Copyright (C) 2004 Hideo Bannai (bannai@ims.u-tokyo.ac.jp)
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "TaggedNumericDataValue.h"

#include <sstream>
#include "Scores.h"

const TaggedNumericDataValue TaggedNumericDataValue::zero =
TaggedNumericDataValue(0, 0.0, 0, 0.0);

TaggedNumericDataValue::TaggedNumericDataValue()
: tagged_cnt(0), untagged_cnt(0), tagged_val(0.0), untagged_val(0.0){
}

TaggedNumericDataValue::~TaggedNumericDataValue()
{}

TaggedNumericDataValue::TaggedNumericDataValue(std::size_t tc, double tv, std::size_t uc, double uv)
: tagged_cnt(tc), untagged_cnt(uc), tagged_val(tv), untagged_val(uv){
}

const std::string TaggedNumericDataValue::get_representation() const {
  std::ostringstream s;
  s << "TaggedNumericDataValue: tagged_count = " << tagged_cnt 
    << " tagged_sum = " << tagged_val
    << " untagged_count = " << untagged_cnt
    << " untagged_sum = " << untagged_val;
  return s.str();
}

void TaggedNumericDataValue::accumulate(const TaggedNumericDataValue & ndv){
  this->tagged_cnt += ndv.tagged_cnt;
  this->tagged_val += ndv.tagged_val;
  this->untagged_cnt += ndv.untagged_cnt;
  this->untagged_val += ndv.untagged_val;
}

void TaggedNumericDataValue::subtract(const TaggedNumericDataValue & ndv){
  this->tagged_cnt -= ndv.tagged_cnt;
  this->tagged_val -= ndv.tagged_val;
  this->untagged_cnt -= ndv.untagged_cnt;
  this->untagged_val -= ndv.untagged_val;
}

void TaggedNumericDataValue::accumulate(std::size_t tc, double tv, std::size_t uc, double uv){
  this->tagged_cnt += tc;
  this->tagged_val += tv;
  this->untagged_cnt += uc;
  this->untagged_val += uv;
}

TaggedNumericDataValueFactory::TaggedNumericDataValueFactory
(const std::vector<double> & v, const std::vector<bool> & tagv)
: weight_vec(v), tag_vec(tagv), total_val(0,0.0,0,0.0){
  std::size_t i;
  for(i = 0; i < v.size(); i++){
    if(tagv[i]){  
      total_val.accumulate(1, weight_vec[i], 0, 0.0);
    } else {
      total_val.accumulate(0, 0.0, 1, weight_vec[i]);
    }
  }
}

TaggedNumericDataValue TaggedNumericDataValueFactory::create_datavalue(std::size_t id) const
{
  TaggedNumericDataValue nv(TaggedNumericDataValue::zero);
  if(tag_vec[id]){
    nv.accumulate(1, weight_vec[id], 0, 0.0);
  } else {
    nv.accumulate(0, 0.0, 1, weight_vec[id]);
  }
  return nv;
}

/*
  a combination of a pattern is expressed by a 
   4 bit integer, 0 to 15.
   p1  p2 | output
   ---------------
   0   0  |  bit 1  1 : (not p1) and (not p2)
   0   1  |  bit 2  2 : (not p1) and p2
   1   0  |  bit 3  4 : p1 and (not p2)
   1   1  |  bit 4  8 : p1 and p2
*/

void calc_combo(const TaggedNumericDataValue & match,
		const TaggedNumericDataValue & total_val,
		std::size_t combo,
		std::size_t & n1, std::size_t & n2,
		double & sum1){
  n1 = 0;
  sum1 = 0.0;
  if(combo & 1){ /*  match Ȃ untagged */
    n1   += total_val.get_untagged_cnt() - match.get_untagged_cnt();
    sum1 += total_val.get_untagged_val() - match.get_untagged_val();
  }
  if(combo & 2){ /*  match  untagged ̐ */
    n1   += match.get_untagged_cnt();
    sum1 += match.get_untagged_val();
  }
  if(combo & 4){ /*  match Ȃ tagged */
    n1   += total_val.get_tagged_cnt() - match.get_tagged_cnt();
    sum1 += total_val.get_tagged_val() - match.get_tagged_val();
  }
  if(combo & 8){ /*  match  tagged */
    n1   += match.get_tagged_cnt();
    sum1 += match.get_tagged_val();
  }

  n2   = total_val.get_tagged_cnt() + total_val.get_untagged_cnt() - n1;
}

double TNDVRankSumTestScore::score(const TaggedNumericDataValue & match,
				   const TaggedNumericDataValue & total_val) const {
  std::size_t n1, n2;
  double sum1;
  calc_combo(match, total_val, combo, n1, n2, sum1);
  // Wilcoxon Rank sum test statistic.
  return(fabs(Scores::ranksumtest(n1, n2, sum1, vc)));
}

std::string TNDVRankSumTestScore::eval_str(const TaggedNumericDataValue & match,
					   const TaggedNumericDataValue & total_val) const {
  std::ostringstream s;
  std::size_t n1, n2;
  double sum1, score;
  calc_combo(match, total_val, combo, n1, n2, sum1);
  score = fabs(Scores::ranksumtest(n1, n2, sum1, vc));
  s << "Combo: " << combo
    << " match = " << n1 << " nomatch = " << n2 
    << " match sum = " << sum1 << " match avg = " << (sum1/n1);
  return s.str();
}
